import * as dntShim from "../../_dnt.shims.js";
import { AsyncQueue, concat } from "../../deps.js";
import { TcpProvider } from "../tcp/tcp_provider.node.js";
import { NotSupportedError } from "../util/errors.js";
import { DecryptLengthDelimitStream, DecryptStream } from "./message_crypto.js";
export class PartnerTcp {
    constructor(conn, appetite, encryptionKey, port) {
        Object.defineProperty(this, "concurrentTransfers", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 128
        });
        Object.defineProperty(this, "payloadThreshold", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 4
        });
        Object.defineProperty(this, "rangeDivision", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: 8
        });
        Object.defineProperty(this, "syncAppetite", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "messageConn", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "encoder", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new dntShim.TextEncoder()
        });
        Object.defineProperty(this, "decoder", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new dntShim.TextDecoder()
        });
        Object.defineProperty(this, "incomingQueue", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new AsyncQueue()
        });
        Object.defineProperty(this, "derivedKey", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "port", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        this.syncAppetite = appetite;
        this.messageConn = conn;
        this.derivedKey = encryptionKey;
        this.port = port;
        const decoder = this.decoder;
        const incomingQueue = this.incomingQueue;
        this.messageConn.readable
            .pipeThrough(new DecryptLengthDelimitStream(encryptionKey))
            .pipeThrough(new DecryptStream(encryptionKey))
            .pipeTo(new dntShim.WritableStream({
            write(chunk) {
                const decoded = decoder.decode(chunk);
                const parsed = JSON.parse(decoded);
                incomingQueue.push(parsed);
            },
            abort() {
                incomingQueue.close();
            },
            close() {
                incomingQueue.close();
            },
        })).catch(() => {
            // The other side may close this connection.
        });
    }
    async sendEvent(event) {
        const encoded = this.encoder.encode(JSON.stringify(event));
        const encrypted = await this.encryptBytes(encoded);
        try {
            await this.messageConn.write(encrypted);
        }
        catch {
            // The connection may have been closed.
        }
    }
    async encryptBytes(bytes) {
        const lenIv = dntShim.crypto.getRandomValues(new Uint8Array(12));
        const bytesIv = dntShim.crypto.getRandomValues(new Uint8Array(12));
        const encryptedBytes = await dntShim.crypto.subtle.encrypt({
            name: "AES-GCM",
            iv: bytesIv,
        }, this.derivedKey, bytes);
        const lengthBytes = new Uint8Array(4);
        const lenView = new DataView(lengthBytes.buffer);
        lenView.setUint32(0, bytesIv.byteLength + encryptedBytes.byteLength);
        const encryptedLength = await dntShim.crypto.subtle.encrypt({
            name: "AES-GCM",
            iv: lenIv,
        }, this.derivedKey, lengthBytes);
        const encryptedMessage = concat(lenIv, new Uint8Array(encryptedLength), bytesIv, new Uint8Array(encryptedBytes));
        return encryptedMessage;
    }
    getEvents() {
        return this.incomingQueue;
    }
    closeConnection() {
        try {
            this.messageConn.close();
        }
        catch {
            // It was closed already (by the other side)
        }
        return Promise.resolve();
    }
    async getDownload(opts) {
        const encoder = new dntShim.TextEncoder();
        const shareAddressBytes = encoder.encode(opts.shareAddress);
        const authorBytes = encoder.encode(opts.doc.author);
        const formatBytes = encoder.encode(opts.doc.format);
        const pathBytes = encoder.encode(opts.doc.path);
        // download ID!
        // authorBytes (59)
        // share len (2)
        // share address (share len)
        // format len (1)
        // format (format len)
        // path len (2)
        // path (path len)
        const transferDescBytes = new Uint8Array(59 +
            1 + shareAddressBytes.byteLength +
            1 + formatBytes.byteLength +
            pathBytes.byteLength);
        let position = 0;
        const transferView = new DataView(transferDescBytes.buffer);
        transferDescBytes.set(authorBytes, position);
        position += authorBytes.byteLength;
        transferView.setUint8(position, shareAddressBytes.byteLength);
        position += 1;
        transferDescBytes.set(shareAddressBytes, position);
        position += shareAddressBytes.byteLength;
        transferView.setUint8(position, formatBytes.byteLength);
        position += 1;
        transferDescBytes.set(formatBytes, position);
        position += formatBytes.byteLength;
        transferDescBytes.set(pathBytes, position);
        const tcpProvider = new TcpProvider();
        const newConn = await tcpProvider.connect({
            port: this.port,
            hostname: this.messageConn.remoteAddr.hostname,
        });
        // Send the byte identifying this connection as an attachment transfer
        const idByte = new Uint8Array(1);
        const idView = new DataView(idByte.buffer);
        // 1 is the ID for an attachment transfer
        idView.setUint8(0, 1);
        await newConn.write(idByte);
        await newConn.write(await this.encryptBytes(transferDescBytes));
        const derivedKey = this.derivedKey;
        let cancelled = false;
        const readable = new dntShim.ReadableStream({
            async start(controller) {
                await newConn.readable
                    // @ts-ignore Node is upset about something here, although it works.
                    .pipeThrough(new DecryptLengthDelimitStream(derivedKey))
                    // And here.
                    // @ts-ignore And here.
                    .pipeThrough(new DecryptStream(derivedKey))
                    .pipeTo(
                // @ts-ignore Node incorrectly thinks Writable streams should have a close method on them.
                new dntShim.WritableStream({
                    write(chunk) {
                        if (!cancelled) {
                            controller.enqueue(chunk);
                        }
                    },
                })).catch((err) => {
                    console.log(err);
                });
                try {
                    newConn.close();
                }
                catch {
                    // Closed by other side
                }
            },
            cancel() {
                cancelled = true;
            },
        });
        return readable;
    }
    handleUploadRequest(_opts) {
        return Promise.resolve(new NotSupportedError());
    }
    handleTransferRequest(source, kind) {
        if (kind === "upload") {
            return Promise.resolve(new NotSupportedError());
        }
        const encryptBytes = this.encryptBytes.bind(this);
        const connWriter = source.writable.getWriter();
        // a writable stream which writes encrypted values to the conn.
        const writable = new dntShim.WritableStream({
            async write(chunk) {
                const encrypted = await encryptBytes(chunk);
                await connWriter.ready;
                await connWriter.write(encrypted);
            },
            abort() {
                try {
                    source.close();
                }
                catch {
                    // Other side closed it
                }
            },
            close() {
                try {
                    source.close();
                }
                catch {
                    // Other side closed it
                }
            },
        });
        return Promise.resolve(writable);
    }
}
